// gtkfeed.cpp

#include "gtkfeed.h"
#include <gtk/gtk.h>

// common

static int get_iter_pos (GtkTreeIter *iter)
{ return GPOINTER_TO_INT (iter->user_data); }

static void set_iter_pos (GtkTreeIter *iter, int pos)
{ iter->user_data = GINT_TO_POINTER (pos); }

static bool is_pos_valid (GtkTreeModel *model, int pos)
{
	if (GTK_IS_FEED_MODEL (model))
		return GTK_FEED_MODEL (model)->feed->getNews (pos) != NULL;
	else
		return GTK_MANAGER_MODEL (model)->manager->getFeed (pos) != NULL;
}

static GtkTreeModelFlags gtk_model_get_flags (GtkTreeModel *model)
{
	return (GtkTreeModelFlags) GTK_TREE_MODEL_LIST_ONLY;
}

static gboolean gtk_model_get_iter (GtkTreeModel *model, GtkTreeIter *iter, GtkTreePath  *path)
{
	int pos = *gtk_tree_path_get_indices (path);
	set_iter_pos (iter, pos);
	return is_pos_valid (model, pos);
}

static GtkTreePath *gtk_model_get_path (GtkTreeModel *model, GtkTreeIter *iter)
{
	int pos = get_iter_pos (iter);
	GtkTreePath *path = gtk_tree_path_new();
	gtk_tree_path_append_index (path, pos);
	return path;
}

static gboolean gtk_model_iter_next (GtkTreeModel *model, GtkTreeIter *iter)
{
	int pos  = get_iter_pos (iter) + 1;
	set_iter_pos (iter, pos);
	return is_pos_valid (model, pos);
}

static gboolean gtk_model_iter_parent (GtkTreeModel *model, GtkTreeIter *iter,
                                       GtkTreeIter  *child)
{ return FALSE; }

static gboolean gtk_model_iter_has_child (GtkTreeModel *model, GtkTreeIter *iter)
{ return FALSE; }

static gint gtk_model_iter_n_children (GtkTreeModel *model, GtkTreeIter *iter)
{ return 0; }

static gboolean gtk_model_iter_nth_child (GtkTreeModel *model, GtkTreeIter *iter,
                                          GtkTreeIter  *parent, gint        n)
{ return FALSE; }

static gboolean gtk_model_iter_children (GtkTreeModel *model, GtkTreeIter *iter,
                                         GtkTreeIter  *parent)
{ return FALSE; }

// GtkManagerModel

struct ManagerListener : public Manager::Listener
{
	ManagerListener (GtkManagerModel *model)
	: mmodel (model)
	{ Manager::get()->addListener (this); }
	~ManagerListener()
	{ Manager::get()->removeListener (this); }

	GtkManagerModel *mmodel;

private:
	virtual void feedStatusChange (Manager *manager, Feed *feed)
	{
		mmodel->notifyPos (manager->getFeedNb (feed));
	}

	// structuralChange are handled by recreating the model
	virtual void startStructuralChange (Manager *manager) {}
	virtual void endStructuralChange (Manager *manager) {}
	// non applicable
	virtual void feedLoading (Manager *manager, Feed *feed) {}
	virtual void feedLoaded (Manager *manager, Feed *feed) {}
	virtual void feedsLoadingProgress (Manager *manager, float fraction) {}
};

static void gtk_manager_model_tree_model_init (GtkTreeModelIface *iface);
static void gtk_manager_model_drag_source_init (GtkTreeDragSourceIface *iface);
static void gtk_manager_model_drag_dest_init (GtkTreeDragDestIface *iface);

G_DEFINE_TYPE_WITH_CODE (GtkManagerModel, gtk_manager_model, G_TYPE_OBJECT,
	G_IMPLEMENT_INTERFACE (GTK_TYPE_TREE_MODEL, gtk_manager_model_tree_model_init)
	G_IMPLEMENT_INTERFACE (GTK_TYPE_TREE_DRAG_SOURCE, gtk_manager_model_drag_source_init)
	G_IMPLEMENT_INTERFACE (GTK_TYPE_TREE_DRAG_DEST, gtk_manager_model_drag_dest_init))

static void gtk_manager_model_init (GtkManagerModel *model)
{
	model->manager = Manager::get();
	model->listener = new ManagerListener (model);
}

static void gtk_manager_model_finalize (GObject *object)
{
	GtkManagerModel *mmodel = GTK_MANAGER_MODEL (object);
	delete mmodel->listener;
	mmodel->listener = 0;
	G_OBJECT_CLASS (gtk_manager_model_parent_class)->finalize (object);
}

static gint gtk_manager_model_get_n_columns (GtkTreeModel *model)
{ return GtkManagerModel::TOTAL_COLS; }

static GType gtk_manager_model_get_column_type (GtkTreeModel *tree_model, gint column)
{
	switch ((GtkManagerModel::Columns) column) {
		case GtkManagerModel::TITLE_COL:
		case GtkManagerModel::TITLE_UNREAD_COL:
		case GtkManagerModel::COLOR_COL:
			return G_TYPE_STRING;
		case GtkManagerModel::WEIGHT_COL:
			return G_TYPE_INT;
		case GtkManagerModel::ICON_COL:
			return G_TYPE_STRING;
		case GtkManagerModel::TOTAL_COLS: return 0;
	}
	return 0;
}

Feed *GtkManagerModel::getFeed (GtkTreeIter *iter)
{
	int pos = get_iter_pos (iter);
	return manager->getFeed (pos);
}

static void gtk_manager_model_get_value (GtkTreeModel *model, GtkTreeIter *iter,
                                         gint column, GValue *value)
{
	GtkManagerModel *mmodel = GTK_MANAGER_MODEL (model);
	Feed *feed = mmodel->getFeed (iter);
	g_value_init (value, gtk_manager_model_get_column_type (model, column));
	switch ((GtkManagerModel::Columns) column) {
		case GtkManagerModel::TITLE_COL: {
			char *str = g_strdup (feed->title().c_str());
			g_value_set_string (value, str);
			break;
		}
		case GtkManagerModel::TITLE_UNREAD_COL: {
			int unread = feed->unreadNb();
			std::string title = feed->title();
			if (title.empty())
				title = feed->_url();
			char *str;
			if (unread) str = g_strdup_printf ("%s (%d)", title.c_str(), unread);
			else str = g_strdup (title.c_str());
			g_value_set_string (value, str);
			break;
		}
		case GtkManagerModel::WEIGHT_COL: {
			int weight = feed->unreadNb() ? PANGO_WEIGHT_BOLD : PANGO_WEIGHT_NORMAL;
			g_value_set_int (value, weight);
			break;
		}
		case GtkManagerModel::COLOR_COL: {
			const char *color = 0;
			if (!feed->errorMsg().empty())
				color = "red";
			else if (feed->loading())
				color = "gray";
			g_value_set_string (value, color);
			break;
		}
		case GtkManagerModel::ICON_COL: {
			const char *stock = GTK_STOCK_FILE; //DIRECTORY;
			if (!feed->errorMsg().empty())
				stock = GTK_STOCK_DIALOG_ERROR;
			else if (feed->loading())
				stock = GTK_STOCK_REFRESH;
			g_value_set_string (value, stock);
			break;
		}
		case GtkManagerModel::TOTAL_COLS: break;
	}
}

void GtkManagerModel::notifyPos (int pos)
{
	GtkTreeModel *model = GTK_TREE_MODEL (this);
	GtkTreeIter iter;
	set_iter_pos (&iter, pos);
	GtkTreePath *path = gtk_model_get_path (model, &iter);
	gtk_tree_model_row_changed (model, path, &iter);
	gtk_tree_path_free (path);
}

static gboolean gtk_manager_model_row_draggable (GtkTreeDragSource *drag_source,
                                                 GtkTreePath       *path)
{ return TRUE; }
  
static gboolean gtk_manager_model_drag_data_delete (GtkTreeDragSource *drag_source,
                                                    GtkTreePath       *path)
{ return FALSE; }

static gboolean gtk_manager_model_drag_data_get (GtkTreeDragSource *drag_source,
                            GtkTreePath *path, GtkSelectionData *selection_data)
{
  if (gtk_tree_set_row_drag_data (selection_data, GTK_TREE_MODEL (drag_source), path))
		return TRUE;
	return FALSE;
}

static gboolean gtk_manager_model_drag_data_received (GtkTreeDragDest *drag_dest,
                             GtkTreePath *dst_path, GtkSelectionData *selection_data)
{
#if 0
	// old behavior would swap src_path and dst_path feed entries
	GtkTreePath *src_path = NULL;
	GtkTreeModel *model = GTK_TREE_MODEL (drag_dest), *src_model;
	gboolean ret = FALSE;

	if (gtk_tree_get_row_drag_data (selection_data, &src_model, &src_path) &&
	    src_model == model) {
		GtkTreeIter src_iter, dst_iter;
		int dst_pos;
		if (!gtk_model_get_iter (model, &dst_iter, dst_path)) {
			dst_pos = get_iter_pos (&dst_iter);
			if (dst_pos < 0) dst_pos = 0;
			else dst_pos--;
		}
		else
			dst_pos = get_iter_pos (&dst_iter);

		if (is_pos_valid (model, dst_pos) &&
		    gtk_model_get_iter (model, &src_iter, src_path)) {
			GtkManagerModel *mmodel = GTK_MANAGER_MODEL (model);
			Feed* feed1 = mmodel->manager->getFeed (get_iter_pos (&src_iter));
			Feed* feed2 = mmodel->manager->getFeed (dst_pos);
			mmodel->manager->swap (feed1, feed2);
			ret = TRUE;
		}
	}
	if (src_path)
		gtk_tree_path_free (src_path);
#endif
	// move feed from src_path to the position of dst_path
	GtkTreePath *src_path = NULL;
	GtkTreeModel *model = GTK_TREE_MODEL (drag_dest), *src_model;
	gboolean ret = FALSE;

	if (gtk_tree_get_row_drag_data (selection_data, &src_model, &src_path) &&
	    src_model == model) {
		int new_pos = *gtk_tree_path_get_indices (dst_path);
		GtkTreeIter src_iter;
		if (gtk_model_get_iter (model, &src_iter, src_path)) {
			GtkManagerModel *mmodel = GTK_MANAGER_MODEL (model);
			Feed* feed = mmodel->manager->getFeed (get_iter_pos (&src_iter));
			mmodel->manager->move (feed, new_pos);
			ret = TRUE;
		}
	}
	if (src_path)
		gtk_tree_path_free (src_path);
	return ret;
}

static gboolean gtk_manager_model_row_drop_possible (GtkTreeDragDest *drag_dest,
                       GtkTreePath *dest_path, GtkSelectionData *selection_data)
{ return TRUE; }

static void gtk_manager_model_drag_source_init (GtkTreeDragSourceIface *iface)
{
	iface->row_draggable = gtk_manager_model_row_draggable;
	iface->drag_data_delete = gtk_manager_model_drag_data_delete;
	iface->drag_data_get = gtk_manager_model_drag_data_get;
}

static void gtk_manager_model_drag_dest_init (GtkTreeDragDestIface *iface)
{
	iface->drag_data_received = gtk_manager_model_drag_data_received;
	iface->row_drop_possible = gtk_manager_model_row_drop_possible;
}

GtkTreeModel *gtk_manager_model_new()
{
	return (GtkTreeModel *) g_object_new (GTK_TYPE_MANAGER_MODEL, NULL);
}

static void gtk_manager_model_class_init (GtkManagerModelClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
	gobject_class->finalize = gtk_manager_model_finalize;
}

static void gtk_manager_model_tree_model_init (GtkTreeModelIface *iface)
{
	iface->get_flags = gtk_model_get_flags;
	iface->get_iter = gtk_model_get_iter;
	iface->get_path = gtk_model_get_path;
	iface->iter_next = gtk_model_iter_next;
	iface->iter_children = gtk_model_iter_children;
	iface->iter_has_child = gtk_model_iter_has_child;
	iface->iter_n_children = gtk_model_iter_n_children;
	iface->iter_nth_child = gtk_model_iter_nth_child;
	iface->iter_parent = gtk_model_iter_parent;
	iface->get_n_columns = gtk_manager_model_get_n_columns;
	iface->get_column_type = gtk_manager_model_get_column_type;
	iface->get_value = gtk_manager_model_get_value;
}

// GtkFeedModel

static void gtk_feed_model_tree_model_init (GtkTreeModelIface *iface);

G_DEFINE_TYPE_WITH_CODE (GtkFeedModel, gtk_feed_model, G_TYPE_OBJECT,
	G_IMPLEMENT_INTERFACE (GTK_TYPE_TREE_MODEL, gtk_feed_model_tree_model_init))

static void gtk_feed_model_init (GtkFeedModel *model)
{
}

static gint gtk_feed_model_get_n_columns (GtkTreeModel *model)
{ return GtkFeedModel::TOTAL_COLS; }

static GType gtk_feed_model_get_column_type (GtkTreeModel *tree_model, gint column)
{
	switch ((GtkFeedModel::Columns) column) {
		case GtkFeedModel::TITLE_COL:
		case GtkFeedModel::DATE_COL:
		case GtkFeedModel::TOOLTIP_COL:
			return G_TYPE_STRING;
		case GtkFeedModel::WEIGHT_COL:
			return G_TYPE_INT;
		case GtkFeedModel::TOTAL_COLS: return 0;
	}
	return 0;
}

News *GtkFeedModel::getNews (GtkTreeIter *iter)
{
	int pos = get_iter_pos (iter);
	return feed->getNews (pos);
}

static void gtk_feed_model_get_value (GtkTreeModel *model, GtkTreeIter *iter,
                                      gint column, GValue *value)
{
	GtkFeedModel *fmodel = GTK_FEED_MODEL (model);
	News *news = fmodel->getNews (iter);
	g_value_init (value, gtk_feed_model_get_column_type (model, column));
	switch ((GtkFeedModel::Columns) column) {
		case GtkFeedModel::TITLE_COL: {
			char *str = g_strdup (news->title().c_str());
			g_value_set_string (value, str);
			break;
		}
		case GtkFeedModel::DATE_COL: {
			char *str = g_strdup (news->date().c_str());
			g_value_set_string (value, str);
			break;
		}
		case GtkFeedModel::TOOLTIP_COL: {
			std::string tooltip;
			tooltip.reserve (2048);
			gchar *title = g_markup_escape_text (news->title().c_str(), -1);
			tooltip += "<b>Title: </b>";
			tooltip += title;
			g_free (title);
			tooltip += "\n\n<b>Date: </b>" + news->date();
			if (!news->author().empty())
				tooltip += "\n<b>Author: </b>" + news->author();
			if (!news->categories().empty())
				tooltip += "\n<b>Categories: </b>" + news->categories();
			g_value_set_string (value, g_strdup (tooltip.c_str()));
			break;
		}
		case GtkFeedModel::WEIGHT_COL: {
			int weight = !news->isRead() ? PANGO_WEIGHT_BOLD : PANGO_WEIGHT_NORMAL;
			g_value_set_int (value, weight);
			break;
		}
		case GtkFeedModel::TOTAL_COLS: break;
	}
}

GtkTreeModel *gtk_feed_model_new (Feed *feed)
{
	GtkFeedModel *model = (GtkFeedModel *) g_object_new (GTK_TYPE_FEED_MODEL, NULL);
	model->feed = feed;
	return (GtkTreeModel *) model;
}

static void gtk_feed_model_class_init (GtkFeedModelClass *klass)
{
}

static void gtk_feed_model_tree_model_init (GtkTreeModelIface *iface)
{
	iface->get_flags = gtk_model_get_flags;
	iface->get_iter = gtk_model_get_iter;
	iface->get_path = gtk_model_get_path;
	iface->iter_next = gtk_model_iter_next;
	iface->iter_children = gtk_model_iter_children;
	iface->iter_has_child = gtk_model_iter_has_child;
	iface->iter_n_children = gtk_model_iter_n_children;
	iface->iter_nth_child = gtk_model_iter_nth_child;
	iface->iter_parent = gtk_model_iter_parent;
	iface->get_n_columns = gtk_feed_model_get_n_columns;
	iface->get_column_type = gtk_feed_model_get_column_type;
	iface->get_value = gtk_feed_model_get_value;
}

